# -*- coding: utf-8 -*-
from __future__ import absolute_import
import os
import six
from six.moves.urllib.parse import urljoin
import six.moves.cPickle as pickle

import parsel
from autopager.htmlutils import get_links
from autopager.model import page_to_features, get_crf, AUTOPAGER_LIMITS
from autopager.storage import Storage


auto = object()


def urls(page, baseurl=auto, direct=True, prev=True, next=True):
    """
    Return a list of pagination URLs extracted form the page.
    When baseurl is None relative URLs are returned; pass baseurl
    to get absolute URLs.

    ``prev``, ``next`` and ``direct`` arguments control whether to return
    'next page', 'previous page' links and links to specific pages.
    By default, all link types are returned.
    """
    return get_shared_autopager().urls(page, baseurl, direct, prev, next)


def select(page, direct=True, prev=True, next=True):
    """
    Return parsel.SelectorList with pagination <a> elements.

    ``prev``, ``next`` and ``direct`` arguments control whether to return
    'next page', 'previous page' links and links to specific pages.
    By default, all link types are returned.
    """
    return get_shared_autopager().select(page, direct, prev, next)


def extract(page, direct=True, prev=True, next=True):
    """
    Return an iterator of (link_type, link) tuples with pagination links.
    link_type is one of "PAGE", "PREV" and "NEXT" constants;
    link is a parsel.Selector instance with pagination <a> element.

    ``prev``, ``next`` and ``direct`` arguments control whether to return
    'next page', 'previous page' links and links to specific pages.
    By default, all link types are returned.
    """
    return list(get_shared_autopager().extract(page, direct, prev, next))


class AutoPager(object):
    DEFAULT_CRF_PATH = os.path.join(os.path.dirname(__file__), 'autopager.crf')

    def __init__(self, path=None, crf=None):
        if crf is not None and path is not None:
            raise ValueError("Pass either path or a CRF instance")
        if crf is None and path is None:
            if not os.path.exists(self.DEFAULT_CRF_PATH):
                raise ValueError("CRF is not trained. If you've installed "
                                 "autopager from source repository you need"
                                 "to create the model yourselves from CLI "
                                 "using ``autopager train`` command.")
            path = self.DEFAULT_CRF_PATH
        if path is not None:
            with open(path, 'rb') as f:
                crf = pickle.load(f)
        self.crf = crf

    def urls(self, page, baseurl=auto, direct=True, prev=True, next=True):
        """
        Return a list of pagination URLs extracted form the page.
        When baseurl is None relative URLs are returned; pass baseurl
        to get absolute URLs.

        ``prev``, ``next`` and ``direct`` arguments control whether to return
        'next page', 'previous page' links and links to specific pages.
        By default, all link types are returned.
        """
        urls = self.select(page, direct, prev, next).xpath("@href").extract()
        if baseurl is auto:
            baseurl = _any2url(page)
        if baseurl is not None:
            urls = [urljoin(baseurl, url) for url in urls]
        return urls

    def select(self, page, direct=True, prev=True, next=True):
        """
        Return parsel.SelectorList with pagination <a> elements.

        ``prev``, ``next`` and ``direct`` arguments control whether to return
        'next page', 'previous page' links and links to specific pages.
        By default, all link types are returned.
        """
        links = self.extract(page, prev=prev, next=next, direct=direct)
        return parsel.SelectorList([x for y, x in links])

    def extract(self, page, direct=True, prev=True, next=True):
        """
        Return an iterator of (link_type, link) tuples with pagination links.
        link_type is one of "PAGE", "PREV" and "NEXT" constants;
        link is a parsel.Selector instance with pagination <a> element.

        ``prev``, ``next`` and ``direct`` arguments control whether to return
        'next page', 'previous page' links and links to specific pages.
        By default, all link types are returned.
        """
        sel = _any2selector(page)
        links = get_links(sel)
        links = links[:AUTOPAGER_LIMITS.max_links]
        xseq = page_to_features(links)
        yseq = self.crf.predict_single(xseq)
        for x, y in zip(links, yseq):
            if direct and y == 'PAGE':
                yield y, x
            if prev and y == 'PREV':
                yield y, x
            if next and y == 'NEXT':
                yield y, x


def _any2selector(data):
    if isinstance(data, parsel.Selector):
        return data

    if isinstance(data, bytes):
        raise ValueError("binary data is not supported")

    if isinstance(data, six.text_type):
        return parsel.Selector(text=data)

    # scrapy.Response support
    try:
        from scrapy.http import TextResponse
    except ImportError:
        pass
    else:
        if isinstance(data, TextResponse):
            return data.selector

    # requests.Response support
    return parsel.Selector(text=data.text)


def _any2url(data):
    return getattr(data, 'url', None)


def train_crf(data_path, out_path):
    """ Train CRF model using default parameters and save it to a file """
    storage = Storage(data_path)
    X_raw, y = storage.get_Xy(validate=False)
    X = [page_to_features(xseq) for xseq in X_raw]
    crf = get_crf()
    crf.fit(X, y)
    with open(out_path, 'wb') as f:
        pickle.dump(crf, f, protocol=2)


_instance = None

def get_shared_autopager():
    """ Return a shared AutoPager instance """
    global _instance
    if _instance is None:
        _instance = AutoPager()
    return _instance



# -*- coding: utf-8 -*-
from __future__ import absolute_import  
import parsel
from autopager.utils import normalize_whitespaces


def get_links_loose(sel):
    """
    Return ``<a>`` elements for a selector.

    If selection is inside <a> element, this parent <a> element is returned.
    If selection is <a> element, it is returned.
    If selection has <a> children, they are returned.
    """
    return parsel.SelectorList(_get_links(sel))


def _get_links(sel):
    for el in sel:
        links_parent = el.xpath('ancestor::a')
        if links_parent:
            for link in links_parent:
                yield link
        else:
            for link in el.css('a'):
                yield link


def get_links(sel):
    return sel.xpath(".//a[@href]")


def get_xseq_yseq(root, selectors, validate=True):
    """
    Extract links from selector ``root`` and assign each link a class label
    if it matches one of the ``selectors`` or "O" if it matches none of them.
    Only ``<a>`` elements with ``href`` attributes are returned.

    ``selectors`` is a ``{label: selector}`` dict.

    When ``validate`` is True this function checks that each selector
    returns non-empty result, and that results don't overlap; in case of
    errors it prints error messages.
    """
    xseq = get_links(root)
    links_extracted = xseq.extract()
    classes = ["O"] * len(xseq)

    for label, selector in selectors.items():
        if not selector:
            continue
        sel_links = set(get_links_loose(root.css(selector)).extract())
        if validate and not sel_links:
            print("error: links not found", label, selector)
        matched_links = set()
        for idx, link in enumerate(links_extracted):
            if link in sel_links:
                if validate and classes[idx] != 'O':
                    print("warning: link is classified in more than 1 classes",
                          link, label)
                matched_links.add(link)
                classes[idx] = label
        if validate and matched_links != sel_links:
            print("Not all links are matched", sel_links - matched_links)

    return xseq, classes


def get_link_text(link):
    """
    Return text of <a> element: either its string contents or the 'alt'
    of <img> tags inside.
    """
    txt = ' '.join(link.xpath('string()').extract())
    if not txt:
        txt = ' '.join(link.xpath('.//img/@alt').extract())
    return txt


def get_link_href(a):
    return a.xpath('@href').extract_first()


def get_text_around_selector_list(sel_list, max_length=150):
    """
    Return a list of (before, after) tuples with text for each element
    in ``sel_list`` parsel.SelectorList. Max text length can be set via
    ``max_length`` argument.
    """
    if not sel_list:
        return []
    doc_root = list(get_selector_root(sel_list[0]).iterancestors())[-1]
    elems = [get_selector_root(sel) for sel in sel_list]
    before, after = get_text_around_elems(doc_root, elems)
    return [(before[el][-max_length:], after[el][:max_length]) for el in elems]


def get_selector_root(sel):
    """ Return lxml Element for a Selector """
    if not hasattr(sel, 'root'):
        # scrapy.Selector, scrapy < 1.1
        return sel._root
    # parsel.Selector or scrapy.Selector, scrapy >= 1.1
    return sel.root


# XXX: this function is copied from formasaurus.
# Time to create an utility library?
def get_text_around_elems(tree, elems):
    """
    Return (before, after) tuple with {elem: text} dicts containing
    text before a specified lxml DOM Element and after it.
    """
    if not elems:
        return {}, {}
    buf = []
    before = {elem: '' for elem in elems}
    after = {elem: '' for elem in elems}

    def flush_buf():
        res = '  '.join([
            normalize_whitespaces(b.strip())
            for b in buf
            if b and b.strip()
        ])
        buf[:] = []
        return res

    def visit(elem):
        if elem in before:
            before[elem] = flush_buf()
            buf.append(elem.tail)
            return
        buf.append(elem.text)
        for child in elem:
            visit(child)
        buf.append(elem.tail)

    visit(tree)

    for prev, next in zip(elems[:-1], elems[1:]):
        after[prev] = before[next]

    after[elems[-1]] = flush_buf()
    return before, after



class Limits(object):
    """
    Limits to avoid exploding memory consumption in some cases
    """
    def __init__(self, max_text_features=80, max_css_features=120, max_links=2500):
        self.max_text_features = max_text_features
        self.max_css_features = max_css_features
        self.max_links = max_links



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import os
import io
import csv
import parsel

from autopager.htmlutils import get_xseq_yseq


DEFAULT_DATA_PATH = os.path.join(os.path.dirname(__file__), 'data')
DEFAULT_LABEL_MAP = {
    'PREV': 'PREV',
    'NEXT': 'NEXT',
    'PAGE': 'PAGE',

    'FIRST': 'PAGE',
    'LAST': 'PAGE',
}


class Storage(object):

    def __init__(self, path=DEFAULT_DATA_PATH, label_map=DEFAULT_LABEL_MAP):
        self.path = path
        self.label_map = label_map

    def get_Xy(self, validate=True):
        X, y = [], []
        for row in self.iter_records():
            html = self._load_html(row)
            selectors = {key: row[key] for key in self.label_map.keys()}
            root = parsel.Selector(html)
            xseq, yseq = get_xseq_yseq(root, selectors, validate=validate)
            yseq = [self.label_map.get(_y, _y) for _y in yseq]
            X.append(xseq)
            y.append(yseq)
        return X, y

    def iter_records(self):
        info_path = os.path.join(self.path, 'data.csv')
        with io.open(info_path, encoding='utf8') as f:
            for row in csv.DictReader(f):
                if row['failed']:
                    continue
                yield row

    def _load_html(self, row):
        data_path = os.path.join(self.path, 'html')
        path = os.path.join(data_path, row['File Name'] + ".html")
        with io.open(path, encoding=row['Encoding']) as f:
            return f.read()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
AutoPager command-line utility.

Usage:
    autopager train <data-path> <modelfile>
    autopager -h | --help
    autopager --version

To train the autopager use "train" command.
"""
from __future__ import absolute_import, print_function
import docopt
import autopager
from autopager.autopager import train_crf


def main():
    args = docopt.docopt(__doc__, version=autopager.__version__)
    if args['train']:
        train_crf(args['<data-path>'], args['<modelfile>'])


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import re
from itertools import chain

import tldextract


def get_domain(url):
    """
    >>> get_domain('example.org')
    'example'
    >>> get_domain('foo.example.co.uk')
    'example'
    """
    return tldextract.extract(url).domain


tokenize = re.compile(r"(?u)\b\w+\b").findall
""" Tokenize text """


_replace_white_spaces = re.compile(r"\s\s+").sub
_replace_newlines = re.compile(r'[\n\r]').sub
def normalize_whitespaces(text):
    """ Replace newlines and whitespaces with a single white space """
    text = _replace_newlines(" ", text)
    return _replace_white_spaces(" ", text)


def ngrams(seq, min_n, max_n):
    """
    Generate min_n to max_n n-grams of elements from a given sequence.
    """
    text_len = len(seq)
    for n in range(min_n, min(max_n + 1, text_len + 1)):
        for i in range(text_len - n + 1):
            yield seq[i: i + n]


def normalize(text):
    """ Default text normalization function """
    return normalize_whitespaces(text.lower())


def ngrams_stripped(text, min_n, max_n):
    for txt in ngrams(text, min_n, max_n):
        stripped = txt.strip()
        if stripped:
            yield stripped


def ngrams_wb(text, min_n, max_n, include_tokens=True):
    """
    Generate character ngrams; they don't span across whitespaces.
    If ``include_tokens`` is True, tokens themselves are included
    if their lenght is less than ``min_n``.

    >>> list(ngrams_wb("I am hungry", 3, 4))
    ['hun', 'ung', 'ngr', 'gry', 'hung', 'ungr', 'ngry', 'I', 'am']
    >>> list(ngrams_wb("I am hungry", 3, 4, include_tokens=False))
    ['hun', 'ung', 'ngr', 'gry', 'hung', 'ungr', 'ngry']
    """
    tokens = text.split()
    for t in tokens:
        for ngram in ngrams_stripped(t, min_n, max_n):
            yield ngram
    if include_tokens:
        for t in tokens:
            if len(t) < min_n:
                yield t


def replace_digits(text, repl='X'):
    """
    >>> replace_digits("hello, 123!")
    'hello, XXX!'
    """
    return re.sub('\d', repl, text)




#!/usr/bin/env python
from setuptools import setup
import re
import os


def get_version():
    fn = os.path.join(os.path.dirname(__file__), "autopager", "__init__.py")
    with open(fn) as f:
        return re.findall("__version__ = '([\d\.\w]+)'", f.read())[0]


setup(
    name='autopager',
    version=get_version(),
    author='Mikhail Korobov',
    author_email='kmike84@gmail.com',
    license='MIT license',
    long_description=open('README.rst').read() + "\n\n" + open('CHANGES.rst').read(),
    description="Detect and classify pagination links on web pages",
    url='https://github.com/TeamHG-Memex/autopager',
    zip_safe=False,
    packages=['autopager'],
    install_requires=[
        "six",
        "w3lib >= 1.13.0",
        "parsel >= 1.0.1",
        "tldextract",
        "docopt",
        "sklearn-crfsuite >= 0.3.3",
    ],
    package_data={
        'autopager': [
            'autopager.crf',
            # 'data/*.csv',
            # 'data/html/*.html'
        ],
    },
    # extras_require={
    #     'with-deps': [
    #         # 'scikit-learn >= 0.17',
    #         # 'scipy',
    #         'sklearn-crfsuite >= 0.3.1',
    #     ],
    # },
    entry_points={
        'console_scripts': ['autopager = autopager.__main__:main']
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
)



# -*- coding: utf-8 -*-
"""

XXX: also tried, but not present in the final model:

* a feature which is 1 if the current link text is a number ``i`` and
  previous link text is ``i-1``;
* extracting number patterns from link text: replacing digits with X and
  characters with C (currently only numbers are replaced);
* replacing numbers with X, not just individual digits;
* character ngrams from raw link href (currently path and query are processed
  separately);
* using query parameter values, not only query parameter names;
* using element id and title attributes;
* all character ngrams or tokens from URL path (currently only a few hardcoded
  features based on path are used);
* string distance between link URL and page URL (Jaro-Winkler on absolute URLs
  was tried; for good results more URL preprocessing is needed);
* distance between URL components (a customized Jaccard distance)


"""
from __future__ import absolute_import

import re
from itertools import islice

import sklearn_crfsuite
from six.moves.urllib.parse import urlsplit, parse_qsl

from autopager.htmlutils import (
    get_link_text,
    get_link_href,
    get_text_around_selector_list,
    get_selector_root,
)
from autopager.limits import Limits
from autopager.utils import normalize, tokenize, ngrams_wb, replace_digits


AUTOPAGER_LIMITS = Limits()


def _elem_attr(elem, attr):
    return normalize(elem.get(attr, ''))


def _num_tokens_feature(text):
    num_tokens = len(tokenize(text))
    if num_tokens > 2:
        num_tokens = '>2'
    else:
        num_tokens = "=%s" % num_tokens
    return num_tokens


def _as_list(generator, limit=None):
    """
    >>> _as_list(ngrams_wb("text", 2, 2), 0)
    []
    >>> _as_list(ngrams_wb("text", 2, 2), 2)
    ['te', 'ex']
    >>> _as_list(ngrams_wb("text", 2, 2))
    ['te', 'ex', 'xt']
    """
    return list(generator if limit is None else islice(generator, 0, limit))


def link_to_features(link):
    text = normalize(get_link_text(link))

    href = get_link_href(link)
    p = urlsplit(href)

    query_parsed = parse_qsl(p.query)
    query_param_names = [k.lower() for k, v in query_parsed]
    query_param_names_ngrams = _as_list(ngrams_wb(
        " ".join([normalize(name) for name in query_param_names]), 3, 5, True
    ))

    elem = get_selector_root(link)
    elem_target = _elem_attr(elem, 'target')
    elem_rel = _elem_attr(elem, 'rel')

    # Classes of link itself and all its children.
    # It is common to have e.g. span elements with fontawesome
    # arrow icon classes inside <a> links.
    self_and_children_classes = ' '.join(link.xpath(".//@class").extract())
    parent_classes = ' '.join(link.xpath('../@class').extract())
    css_classes = normalize(parent_classes + ' ' + self_and_children_classes)

    return {
        'bias': 3.0,
        'isdigit': text.isdigit(),
        'isalpha': text.isalpha(),
        'elem-target': elem_target,
        'elem-rel': elem_rel,
        'num-tokens%s' % _num_tokens_feature(text): 1.0,

        'text': _as_list(ngrams_wb(replace_digits(text), 2, 5),
                         AUTOPAGER_LIMITS.max_text_features),
        'text-exact': replace_digits(text.strip()[:20].strip()),
        'class': _as_list(ngrams_wb(css_classes, 4, 5),
                          AUTOPAGER_LIMITS.max_css_features),
        'query': query_param_names_ngrams,

        'path-has-page': 'page' in p.path.lower(),
        'path-has-pageXX': re.search(r'[/-](?:p|page\w?)/?\d+', p.path.lower()) is not None,
        'path-has-number': any(part.isdigit() for part in p.path.split('/')),

        'href-has-year': re.search('20\d\d', href) is not None,
    }

    # Unused features
    # ===============
    #
    # href_ngrams = ngrams_striped(p.path + '?' + p.query, 5, 5)
    # query_param_patterns = [
    #     "%s=%s" % (k.lower(), number_pattern(v, ratio=0.0).strip())
    #     for k, v in query_parsed
    # ]
    # path_tokens = tokenize(number_pattern(p.path.lower(), 0.1))
    # path_ngrams = ngrams_striped(p.path, 4, 4)
    # query_ngrams = ngrams_striped(p.query, 4, 4)
    #
    # path_tokens = ngrams_wb(replace_digits(p.path.lower().replace('/', ' ')), 3, 5)
    # elem_css_class = _elem_attr(elem, 'class')
    # elem_id = _elem_attr(elem, 'id')
    # elem_title = _elem_attr(elem, 'title')
    #
    # path = path_tokens
    # bad_href = "javascript://" in _href or p.fragment and (not p.query or p.path)
    # title = ngrams_wb(replace_digits(normalize(elem_title)), 4, 5)
    #
    # 'css-class-ngrams': ngrams_striped(link_classes, 4, 5),
    # 'css-parent-class-ngrams': ngrams_striped(parent_classes, 4, 5),
    # 'query_param_name': query_param_names,
    # 'query_param_pattern': query_param_patterns,
    # 'path_tokens': path_tokens,
    # 'text_pattern': normalize(text_pattern),
    # 'text_ngrams': text_ngrams,
    #
    # 'id': ngrams_wb(elem_id, 4, 4),
    # 'id-class-ngrams': ngrams_striped(elem_css_class + ' ' + elem_id, 5, 5),
    # 'id': tokenize(elem_id),


def page_to_features(xseq):
    features = [link_to_features(a) for a in xseq]
    around = get_text_around_selector_list(xseq, max_length=15)

    # weight is less than 1 because there is a lot of duplicate information
    # in these ngrams and so we want to regularize them stronger
    # (as if they are a single feature, not many features)
    k = 0.2
    for feat, (before, after) in zip(features, around):
        feat['text-before'] = {n: k for n in _as_list(ngrams_wb(normalize(before), 5, 5))}
        feat['text-after'] = {n: k for n in _as_list(ngrams_wb(normalize(after), 5, 5))}
    return features


def get_crf(**kwargs):
    params = dict(
        algorithm='lbfgs',
        c1=0.001,
        c2=0.05,
        max_iterations=100,
        all_possible_transitions=True,
        verbose=False,
    )
    params.update(kwargs)
    return sklearn_crfsuite.CRF(**params)
